import numpy as np

from snnw.nn.weight_init import kaiming


class Layer:
    # These layers are fully connected by default
    def __init__(self, layer_type, in_size, out_size, activation):
        # only type supported is fully connected
        assert (layer_type == "fully_connected"), "only fully connected layers are supported"
        self.layer_type = "fully_connected"
        # in_size is number of neuron in previous layer
        self.in_size = in_size
        # out_size is number of neurons in current layer
        self.out_size = out_size
        # activation function name
        self.activation = activation
        # make random weight array
        self.weights = kaiming(in_size, out_size)
        # make random bias array
        self.biases = np.zeros(shape=(1, self.out_size))
        # make output array
        self.output = None



import os

import numpy as np

from snnw.nn.model import Model
from snnw.nn.config import load


def train(model_dir, config_path, image_path, label_path, steps=60000, learning_rate=5e-4):
    model_config = load(os.path.realpath(config_path))

    train_images = np.load(os.path.realpath(image_path))
    train_labels = np.load(os.path.realpath(label_path))

    train_model = Model(x_input=train_images, y_true=train_labels, model_config=model_config)
    train_model.train(steps=steps, learning_rate=learning_rate, model_dir=model_dir)


def evaluate(model_dir, config_path, image_path, label_path):
    model_config = load(os.path.realpath(config_path))

    test_images = np.load(os.path.realpath(image_path))
    test_labels = np.load(os.path.realpath(label_path))

    test_model = Model(x_input=test_images, y_true=test_labels, model_config=model_config)
    test_model.evaluate(model_dir=model_dir)




# Begin 'get'

config_1 = """# 0
[fully_connected]
in_size=784
out_size=1000
activation=relu

# 1
[fully_connected]
in_size=1000
out_size=1000
activation=relu

# 2
[fully_connected]
in_size=1000
out_size=10
activation=softmax
"""

config_2 = """# 0
[fully_connected]
in_size=784
out_size=1000
activation=tanh

# 1
[fully_connected]
in_size=1000
out_size=1000
activation=tanh

# 2
[fully_connected]
in_size=1000
out_size=10
activation=softmax
"""


def get(config_path, num):
    assert (num in [1, 2]), "Your config number is not in range."
    config = ""
    if num == 1:
        config = config_1
    elif num == 2:
        config = config_2
    with open(config_path, 'w') as f:
        f.write(config)


# Begin 'load'


def load(config_path):
    with open(config_path, 'r') as f:
        # each layer config is separated by an empty new line
        layers = f.read().split('\n\n')
    layer_configs = []
    for layer in layers:
        layer_config = {}
        layer = layer.splitlines()
        assert (len(layer) == 5), "check that your config format matches the examples given"
        for line in layer:
            in_size = 'in_size='
            out_size = 'out_size='
            activation = 'activation='
            if line[0] == '#':
                continue
            elif line[0] == '[' and line[-1] == ']':
                layer_config['layer_type'] = line[1:-1]
            elif line[:len(in_size)] == in_size:
                layer_config['in_size'] = int(line[len(in_size):])
            elif line[:len(out_size)] == out_size:
                layer_config['out_size'] = int(line[len(out_size):])
            elif line[:len(activation)] == activation:
                layer_config['activation'] = line[len(activation):]
        layer_configs.append(layer_config)
    return layer_configs



import os
import subprocess

url = "http://yann.lecun.com/exdb/mnist/"

files = [
    "train-images-idx3-ubyte.gz",
    "train-labels-idx1-ubyte.gz",
    "t10k-images-idx3-ubyte.gz",
    "t10k-labels-idx1-ubyte.gz",
]


def raw(raw_dir):
    raw_dir = os.path.realpath(raw_dir)
    if not os.path.exists(raw_dir):
        os.makedirs(raw_dir)
    cwd = os.getcwd()
    os.chdir(raw_dir)
    for f in files:
        curl_out = subprocess.Popen('curl {} -O'.format(url + f).split(' '),
                                    stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        curl_stdout, curl_stderr = curl_out.communicate()
        print(curl_stdout.decode('utf-8'))
        if curl_stderr is not None:
            print(curl_stderr.decode('utf-8'))

        gunzip_out = subprocess.Popen('gunzip {}'.format(f).split(' '),
                                      stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        gunzip_stdout, gunzip_stderr = gunzip_out.communicate()
        print(gunzip_stdout.decode('utf-8'))
        if gunzip_stderr is not None:
            print(gunzip_stderr.decode('utf-8'))
    os.chdir(cwd)



import os

import numpy as np
from tqdm import tqdm

from snnw.nn.loss import cross_entropy
from snnw.nn.activation import sigmoid, softmax, relu, tanh
from snnw.nn.layer import Layer

name2function = {
    'relu': relu.function,
    'sigmoid': sigmoid.function,
    'softmax': softmax.function,
    'tanh': tanh.function,
}

name2derivative = {
    'relu': relu.derivative,
    'sigmoid': sigmoid.derivative,
    # 'softmax': softmax.derivative,
    'tanh': tanh.derivative,
}


class Model:
    def __init__(self, x_input, y_true, model_config, loss_class=None):
        # x_in = [# examples x len(x)] numpy array
        self.x_input = x_input

        # y = [# examples x 1] numpy array
        self.y_true = y_true

        # the format of config is a list of dictionaries of
        # {layer_type: ..., in_size: ..., out_size: ..., activation: ...}
        # note: layers must be fully connected

        # first check that the dimensions of everything match up
        dim_error_str = "Model config dimensions do not match. (in_size of layer i != out_size of layer i-1)"
        assert (self.x_input.shape[1] == model_config[0]['in_size']), dim_error_str
        for i in range(1, len(model_config)):
            assert (model_config[i-1]['out_size'] == model_config[i]['in_size']), dim_error_str

        # check that the last layer has activation softmax
        activation_error_str = 'Only softmax and cross-entropy are supported for the last layer and loss, respectively'
        assert (model_config[-1]['activation'] == 'softmax'), activation_error_str

        # initialize model's list of layers
        self.layers = []
        for layer_config in model_config:
            layer = Layer(layer_type=layer_config['layer_type'],
                          in_size=layer_config['in_size'],
                          out_size=layer_config['out_size'],
                          activation=layer_config['activation'])
            self.layers.append(layer)

        # only loss supported is cross-entropy
        self.loss_class = cross_entropy

    def forward(self, x_in, y_gt):
        # x_in shape: [1 x len(x_in)]
        y = x_in
        for layer in self.layers:
            x = np.matmul(y, layer.weights) + layer.biases
            y = name2function[layer.activation](x)
            layer.output = y
        # this should get logits array
        y_pred = y
        E = self.loss_class.function(y_pred=y_pred, y_gt=y_gt)
        return y_pred, E

    def forward_and_backward(self, x_in, y_gt, learning_rate):
        # get logits from forward pass
        y_pred, E = self.forward(x_in, y_gt)

        for i in range(len(self.layers) - 1, -1, -1):
            # shape: [1 x len(logits)]
            # ??? What if the derivative depends on the actual coordinates ???
            if i == len(self.layers) - 1:
                # special case for loss layer
                # We only have Cross-Entropy loss, so just calculate derivative of loss with respect to x before softmax
                dEdx = y_pred * np.sum(y_gt) - y_gt
            else:
                dEdy = np.matmul(dEdx, self.layers[i + 1].weights.T)
                # shape: [1 x len(logits)]
                # Hadamard of dy/dx and dE/dy = dE/dx
                dEdx = name2derivative[self.layers[i].activation](self.layers[i].output) * dEdy

            # shape: [len(prev_layer) x len(logits)]
            if i == 0:
                # special case for input layer
                dEdw = np.matmul(x_in.T, dEdx)
            else:
                dEdw = np.matmul(self.layers[i - 1].output.T, dEdx)

            # shape: [1 x len(logits)]
            dEdb = dEdx

            # adjust the weights and biases for current layer
            self.layers[i].weights = self.layers[i].weights - learning_rate * dEdw
            self.layers[i].biases = self.layers[i].biases - learning_rate * dEdb

        return y_pred, E

    def train(self, steps, learning_rate, model_dir):
        for i in tqdm(range(steps)):
            # stochastic gradient descent
            idx = int(self.x_input.shape[0] * np.random.uniform())
            x_in = self.x_input[idx:idx+1, :]
            y_gt = self.y_true[idx:idx+1, :]
            y_pred, loss = self.forward_and_backward(x_in, y_gt, learning_rate)
            if i % 100 == 0:
                print('Train step {}: \t\tLoss = {:8}'.format(i, loss))

        print('Finished training. Saving weights and biases...')

        weight_arr = []
        bias_arr = []
        for layer in tqdm(self.layers):
            weight_arr.append(layer.weights)
            bias_arr.append(layer.biases)
        # to access the weights of layer 3: weight_arr['arr_3']
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)
        np.savez(os.path.join(model_dir, 'weights.npz'), *weight_arr)
        np.savez(os.path.join(model_dir, 'biases.npz'), *bias_arr)

        print('Finished saving weights and biases.')
        return

    def evaluate(self, model_dir):
        # load weights and biases
        weights_dict = np.load(os.path.join(model_dir, 'weights.npz'))
        biases_dict = np.load(os.path.join(model_dir, 'biases.npz'))
        assert (len(weights_dict) == len(self.layers)), "saved weights' shape != model weights' shape"
        assert (len(biases_dict) == len(self.layers)), "saved biases' shape != model biases' shape"

        # set model's weights and biases to the loaded weights and biases
        for i in range(len(self.layers)):
            layer = self.layers[i]
            key = 'arr_{}'.format(i)
            layer.weights = weights_dict[key]
            layer.biases = biases_dict[key]

        # evaluate for each test image
        correct = 0
        for i in range(self.x_input.shape[0]):
            x_in = self.x_input[i:i+1, :]
            y_gt = self.y_true[i:i+1, :]
            y_pred, E = self.forward(x_in, y_gt)
            if np.argmax(y_pred) == np.argmax(self.y_true[i, :]):
                correct += 1
            if i % 100 == 0:
                print('Evaluate step {}: \t\tLoss = {:8} \t\tAccuracy = {:8}'.format(i, E, correct / float(i + 1)))
        print('Finished evaluating. Final accuracy: {}'.format(correct/float(self.x_input.shape[0])))



import os
import struct

from array import array

import numpy as np
import png
from PIL import Image
from tqdm import tqdm


# Begin 'raw_to_png'


def read(dataset="train", path="."):
    if dataset is "train":
        fname_img = os.path.join(path, 'train-images-idx3-ubyte')
        fname_lbl = os.path.join(path, 'train-labels-idx1-ubyte')
    elif dataset is "test":
        fname_img = os.path.join(path, 't10k-images-idx3-ubyte')
        fname_lbl = os.path.join(path, 't10k-labels-idx1-ubyte')
    else:
        raise ValueError("dataset must be 'train' or 'test'")

    flbl = open(fname_lbl, 'rb')
    magic_nr, size = struct.unpack(">II", flbl.read(8))
    lbl = array("b", flbl.read())
    flbl.close()

    fimg = open(fname_img, 'rb')
    magic_nr, size, rows, cols = struct.unpack(">IIII", fimg.read(16))
    img = array("B", fimg.read())
    fimg.close()

    return lbl, img, size, rows, cols


def write_dataset(labels, data, size, rows, cols, output_path, dataset):
    output_dir = os.path.join(output_path, dataset)
    # create output directories
    output_dirs = [os.path.join(output_dir, str(i)) for i in range(10)]
    for dir in output_dirs:
        if not os.path.exists(dir):
            os.makedirs(dir)

    txt_path = os.path.join(output_path, dataset + ".txt")

    # erase .txt file if it exists
    open(txt_path, "w").close()

    # write data
    i = 0
    for label in tqdm(labels):
        output_filename = os.path.join(output_dirs[label], str(i) + ".png")
        with open(output_filename, "wb") as h:
            w = png.Writer(cols, rows, greyscale=True)
            data_i = [data[(i*rows*cols + j*cols): (i*rows*cols + (j+1)*cols)] for j in range(rows)]
            w.write(h, data_i)
        with open(txt_path, "a") as f:
            f.write(output_filename + "\n")
        i += 1


def raw_to_png(raw_dir, png_dir):
    raw_dir = os.path.abspath(raw_dir)
    png_dir = os.path.abspath(png_dir)

    if not os.path.exists(png_dir):
        os.makedirs(png_dir)

    for dataset in ["train", "test"]:
        print("Writing {} dataset".format(dataset))
        labels, data, size, rows, cols = read(dataset, raw_dir)
        write_dataset(labels, data, size, rows, cols, png_dir, dataset)


# Begin 'png_to_npy'


def mnist_png_to_npy(split, png_dir, npy_dir):
    with open(os.path.join(png_dir, split + '.txt'), 'r') as f:
        paths = f.read().splitlines()
    # number of labels
    labels = 0
    # doesn't matter what path we use to get the split directory name
    split_dir = os.path.dirname(os.path.dirname(paths[0]))
    for _, dirnames, _ in os.walk(split_dir):
        labels += len(dirnames)

    np_images = []
    np_labels = []
    for path in tqdm(paths):
        # np_image shape: [1 x 784]
        np_image = np.array(Image.open(path)).flatten()
        # remember to divide by 256
        np_image = np_image / 256.0
        np_images.append(np_image)

        # get label from folder name of file
        label = int(os.path.basename(os.path.dirname(path)))

        # np_label shape: [1 x #labels]
        np_label = np.zeros(shape=labels)
        np_label[label] = 1.0
        np_labels.append(np_label)
    np.save(os.path.join(npy_dir, split + '-images.npy'), np_images)
    np.save(os.path.join(npy_dir, split + '-labels.npy'), np_labels)


def png_to_npy(png_dir, npy_dir):
    png_dir = os.path.realpath(png_dir)
    npy_dir = os.path.realpath(npy_dir)
    if not os.path.exists(npy_dir):
        os.makedirs(npy_dir)
    splits = ['train', 'test']
    for split in splits:
        print('Current split: ' + split)
        mnist_png_to_npy(split, png_dir, npy_dir)



import numpy as np


def function(x):
    return np.where(x >= 0.0,
                    1.0 / (1.0 + np.exp(-x)),
                    np.exp(x) / (1.0 + np.exp(x)))


def derivative(x):
    sigmoid = function(x)
    return sigmoid * (1 - sigmoid)



import numpy as np


def function(y_pred, y_gt):
    # shape y_pred [1 x _]
    # y_true is [1 x _]
    error_msg = 'y_pred fed into cross entropy is not a probability array, try adjusting your learning rate'
    assert (abs(np.sum(y_pred) - 1.0) < 1e-5), error_msg
    log_likelihood = -np.log(y_pred)
    loss = np.sum(log_likelihood * y_gt) / np.sum(y_gt)
    return loss


def derivative(y_pred, y_true):
    pass
